from typing import TypeVar

Engine = TypeVar(
    "Engine",
)

DBConnection = TypeVar(
    "DBConnection",
)

EngineConnectionPool = TypeVar(
    "EngineConnectionPool",
)

EngineTransaction = TypeVar(
    "EngineTransaction",
)
